from __future__ import annotations
from dataclasses import dataclass, field
from math import ceil
from .errors import vert
from .serialization import serialize_part, deserialize_part
from .vm import (
    OpCode,
    VirtualMachine,
    get_empty_hash,
    hash_leaf,
    hash_node,
    compile,
)


def get_bitpath(data: bytes) -> list[bool]:
    """Get the bitpath for arbitrary bytes data."""
    bitpath = []
    for i in range(len(data)):
        bitpath.extend([
            0b10000000 & data[i] != 0,
            0b01000000 & data[i] != 0,
            0b00100000 & data[i] != 0,
            0b00010000 & data[i] != 0,
            0b00001000 & data[i] != 0,
            0b00000100 & data[i] != 0,
            0b00000010 & data[i] != 0,
            0b00000001 & data[i] != 0,
        ])
    return bitpath

def calculate_intersection(hash1: bytes|int, hash2: bytes|int) -> int:
    """Calculate the intersection point of two hashes. Since the SMT is
        a balanced binary tree, the intersection point is the first bit
        that is different between the two hashes.
    """
    idx1 = int.from_bytes(hash1, 'big') if isinstance(hash1, bytes) else hash1
    idx2 = int.from_bytes(hash2, 'big') if isinstance(hash2, bytes) else hash2
    if idx1 == idx2:
        return 0
    return (idx1^idx2).bit_length()


@dataclass
class SparseSubTree:
    leaf: bytes = field()
    level: int = field()
    _root: bytes = field(default=None)
    _bitpath: list[bool] = field(default=None)

    @property
    def leaf_hash(self) -> bytes:
        return hash_leaf(self.leaf)

    def get_bitpath(self, min_level: int = 8) -> list[bool]:
        """Gets the path from leaf to root for the leaf. Assumes
            big-endian byte order in the leaf hash. Returns a list of
            booleans, where each boolean indicates whether the node is
            on the left (False) or right (True), where the most
            significant bit is the leftmost (lowest index) bit. In a
            sense, the leaf hash can be thought of as one long list of
            bits comprising the integer index of the leaf among the leaf
            positions in a sparse Merkle tree. This is then used to
            calculate the intersection point of two sparse subtrees to
            join them into a single subtree.
        """
        if self._bitpath is None:
            leaf_hash = hash_leaf(self.leaf)
            bitpath = []
            for i in range(max(ceil(self.level/8), ceil(min_level/8), 1)):
                bitpath.extend([
                    0b10000000 & leaf_hash[i] != 0,
                    0b01000000 & leaf_hash[i] != 0,
                    0b00100000 & leaf_hash[i] != 0,
                    0b00010000 & leaf_hash[i] != 0,
                    0b00001000 & leaf_hash[i] != 0,
                    0b00000100 & leaf_hash[i] != 0,
                    0b00000010 & leaf_hash[i] != 0,
                    0b00000001 & leaf_hash[i] != 0,
                ])
            self._bitpath = bitpath
        return self._bitpath

    def intersection_level(self, other: SparseSubTree) -> int:
        """Calculate the intersection level of two sparse subtrees."""
        hash1 = hash_leaf(self.leaf)
        hash2 = hash_leaf(other.leaf)
        return calculate_intersection(hash1, hash2)

    def prove(self) -> list[tuple[bytes|OpCode|int,]]:
        """Create an inclusion proof for this SpareSubTree."""
        leaf_hash = hash_leaf(self.leaf)
        bitpath = self.get_bitpath()

        proof = [
            (OpCode.set_hsize, len(leaf_hash)),
        ]
        accumulated = leaf_hash
        if not bitpath[0]:
            proof.extend([
                (OpCode.load_left, self.leaf),
                (OpCode.hash_leaf_left,),
                (OpCode.load_empty_right, 0)
            ])
            accumulated = hash_node(accumulated, get_empty_hash(0))
        else:
            proof.extend([
                (OpCode.load_right, self.leaf),
                (OpCode.hash_leaf_right,),
                (OpCode.load_empty_left, 0)
            ])
            accumulated = hash_node(get_empty_hash(0), accumulated)

        for i in range(1, self.level):
            if bitpath[i]:
                proof.extend([
                    (OpCode.hash_right,),
                    (OpCode.load_empty_left, i)
                ])
                accumulated = hash_node(get_empty_hash(i), accumulated)
            else:
                proof.extend([
                    (OpCode.hash_left,),
                    (OpCode.load_empty_right, i)
                ])
                accumulated = hash_node(accumulated, get_empty_hash(i))

        proof.append(
            (OpCode.hash_final_hsize, accumulated)
        )

        return proof

    def path(self) -> list[bytes]:
        """Return a simple/compact Merkle audit path for this sparse
            subtree using a hash_to_level_ op.
        """
        path = [
            bytes(OpCode.load_left) + len(self.leaf).to_bytes(2, 'big') +
            self.leaf
        ]
        path.append(bytes(OpCode.hash_leaf_left))
        path.append(bytes(OpCode.set_path_auto))
        path.append(bytes(OpCode.hash_to_level_path) + b'\x00' +
                     self.level.to_bytes(2, 'big'))
        return path

    def root(self) -> bytes:
        """Returns the calculated root."""
        if self._root is None:
            self._root = self.prove()[-1][1]
        return self._root

    def pack(self) -> bytes:
        """Pack the SparseSubTree into bytes."""
        return serialize_part([
            self.level,
            self.leaf,
        ])

    @classmethod
    def unpack(cls, data: bytes, /, *, inject: dict = {}) -> SparseSubTree:
        """Unpack a SpareSubTree from bytes."""
        depedencies = {**globals(), **inject}
        level, leaf = deserialize_part(data, inject=depedencies)
        return cls(
            leaf=leaf,
            level=level,
        )

    def __repr__(self) -> str:
        return f"SparseSubTree(level={self.level}, leaf={self.leaf.hex()})"

    def __eq__(self, other) -> bool:
        return isinstance(other, SparseSubTree) and self.leaf == other.leaf and \
            self.level == other.level


@dataclass
class SparseTree:
    subtrees: list[SparseSubTree] = field(default_factory=list)
    treemap: dict[int, list[int, int]] = field(default_factory=dict)
    root: bytes = field(default=b'')

    @staticmethod
    def hash_up_to(node: bytes, bitpath: list[bool], start_level: int,
                   end_level: int) -> bytes:
        """Hash up to the proper level."""
        # print(f"hash_up_to {node.hex()=} {len(bitpath)=} {start_level=} {end_level=}")
        for i in range(start_level, end_level):
            if not bitpath[i]:
                node = hash_node(node, get_empty_hash(i))
            else:
                node = hash_node(get_empty_hash(i), node)
        return node

    @classmethod
    def from_leaves(cls, leaves: list[bytes]) -> SparseTree:
        lhf = len(hash_leaf(leaves[0]))
        subtrees: list[SparseSubTree] = [
            SparseSubTree(leaf, lhf * 8) for leaf in leaves
        ]
        intersections: list[tuple[int,int,int],] = []

        for i in range(len(subtrees)):
            for j in range(len(subtrees)):
                if i <= j:
                    break
                level = subtrees[i].intersection_level(subtrees[j])
                intersections.append((i, j, level))

        intersections.sort(key=lambda i: i[2])

        treemap: dict[int, list[int, int]] = {}
        for nt in intersections:
            i, j, level = nt
            if level < subtrees[i].level:
                subtrees[i].level = level
            if level < subtrees[j].level:
                subtrees[j].level = level

            if i not in treemap:
                treemap[i] = [level, j]
            if level < treemap[i][0]:
                treemap[i] = [level, j]
            if j not in treemap:
                treemap[j] = [level, i]
            if level < treemap[j][0]:
                treemap[j] = [level, i]

        accumulators = {
            i: {
                'level': subtrees[i].level,
                'value': subtrees[i].root(),
            }
            for i,_ in treemap.items()
        }

        for nt in intersections:
            i, j, level = nt
            if treemap[i] != [level, j]:
                continue
            if accumulators[i]['level'] < level:
                accumulators[i]['value'] = cls.hash_up_to(
                    accumulators[i]['value'],
                    subtrees[i].get_bitpath(),
                    accumulators[i]['level'],
                    level
                )
            if accumulators[j]['level'] < level:
                accumulators[j]['value'] = cls.hash_up_to(
                    accumulators[j]['value'],
                    subtrees[j].get_bitpath(),
                    accumulators[j]['level'],
                    level
                )
            go_left = not subtrees[i].get_bitpath()[subtrees[i].level]
            if go_left:
                accumulators[i]['value'] = hash_node(
                    accumulators[i]['value'],
                    accumulators[j]['value'],
                )
            else:
                accumulators[i]['value'] = hash_node(
                    accumulators[j]['value'],
                    accumulators[i]['value'],
                )

        accumulators = [(k, v) for k,v in accumulators.items()]
        accumulators.sort(key=lambda a: a[1]['level'], reverse=True)

        root = cls.hash_up_to(
            accumulators[-1][1]['value'],
            subtrees[accumulators[-1][0]].get_bitpath(min_level=lhf*8),
            accumulators[-1][1]['level'],
            lhf*8,
        )

        return cls(
            subtrees=subtrees,
            treemap=treemap,
            root=root
        )

    def prove(self, leaf: bytes) -> bytes:
        """Prove leaf inclusion. Returns a proof as VM bytecode."""
        vert(leaf in [t.leaf for t in self.subtrees], 'unrecognized leaf')

        # Get the target subtree and its index
        subtree_index = next(i for i, t in enumerate(self.subtrees) if t.leaf == leaf)
        subtree = self.subtrees[subtree_index]

        # Get the full path from leaf to root
        bitpath = subtree.get_bitpath()

        # Start with the leaf's proof as a subroutine based on path direction
        initial_proof = compile(*subtree.prove())
        proof = []

        if not bitpath[subtree.level]:
            proof.extend([
                (OpCode.subroutine_left, initial_proof)
            ])
        else:
            proof.extend([
                (OpCode.subroutine_right, initial_proof)
            ])

        # Find all intersections involving this subtree, sorted by level
        relevant_intersections = []
        for i, other_index in self.treemap.items():
            if i == subtree_index:
                level, other_subtree = other_index
                relevant_intersections.append((level, other_subtree))
            elif other_index[1] == subtree_index:
                level = other_index[0]
                relevant_intersections.append((level, i))

        relevant_intersections.sort(key=lambda x: x[0])

        # For each intersection level, add the sibling hash to the proof
        for level, other_index in relevant_intersections:
            other_subtree: SparseSubTree = self.subtrees[other_index]

            # Hash up the other subtree to the intersection level
            if other_subtree.level < level:
                other_value = self.hash_up_to(
                    other_value,
                    other_subtree.get_bitpath(),
                    other_subtree.level,
                    level
                )
            else:
                other_value = other_subtree.root()

            # Add the sibling hash to the proof based on path direction
            if not bitpath[level]:
                proof.extend([
                    (OpCode.load_right, other_value),
                    (OpCode.hash_right,)
                ])
            else:
                proof.extend([
                    (OpCode.load_left, other_value),
                    (OpCode.hash_left,)
                ])

        # Add final root verification
        proof.append((OpCode.hash_final_hsize, self.root))

        return compile(*proof)

    def prove_excluded(self, leaf: bytes) -> bytes:
        """Exclusion proofs: find first intersection with bitpath path;
            hash an empty node up to the intersection; then hash up to
            the root. The verification will show that the position the
            leaf would have taken is empty, and therefore the leaf is
            not included in the tree. Returns a proof as VM bytecode.
        """
        ...

    def pack(self) -> bytes:
        """Serialize to bytes."""
        return serialize_part(self.subtrees)

    @classmethod
    def unpack(cls, data: bytes, /, *, inject: dict = {}) -> SparseTree:
        """Deserialize from bytes."""
        dependencies = {**globals(), **inject}
        subtrees = deserialize_part(data, inject=dependencies)
        return cls(
            subtrees=subtrees,
        )

    def __eq__(self, other) -> bool:
        return isinstance(other, SparseTree) and self.root == other.root



import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import merkleasy
from merkleasy import classes
from merkleasy import errors
from merkleasy import interfaces
from merkleasy import serialization
from merkleasy import sparse
from merkleasy import vm
from merkleasy import xorhashtree


class ImplementationError(BaseException):
    """An error that occurs if a function is not properly implemented."""
    ...

class SecurityError(BaseException):
    """An error that occurs if a security check fails."""
    ...

def eruces(condition: bool, error_message: str) -> None:
    """Raises a SecurityError with the given error_message if the
        condition is False. Replacement for assert statements and
        AssertionError during proof verification.
    """
    if not condition:
        raise SecurityError(error_message)

def tert(condition: bool, message: str = '') -> None:
    """Replacement for assert preconditions. Raises TypeError with the
        given message if the condition check fails.
    """
    if condition:
        return
    raise TypeError(message)

def vert(condition: bool, message: str = '') -> None:
    """Replacement for assert preconditions. Raises ValueError with the
        given message if the condition check fails.
    """
    if condition:
        return
    raise ValueError(message)

def yert(condition: bool, message: str = '') -> None:
    """Replacement for assert preconditions. Raises SyntaxError with the
        given message if the condition check fails.
    """
    if condition:
        return
    raise SyntaxError(message)



from context import classes, errors, vm, interfaces
from hashlib import sha256, sha3_256
import unittest


class TestVM(unittest.TestCase):
    def setUp(self) -> None:
        self.original_hash_fn = vm.get_hash_function()
        while len(vm._EMPTY_HASHES):
            vm._EMPTY_HASHES.pop()
        self.root = sha256(b'root').digest()
        return super().setUp()

    def tearDown(self) -> None:
        vm.set_hash_function(self.original_hash_fn)
        while len(vm._EMPTY_HASHES):
            vm._EMPTY_HASHES.pop()
        return super().tearDown()

    @classmethod
    def setUpClass(cls) -> None:
        cls.leaves = [i.to_bytes(1, 'big') for i in range(10)]
        return super().setUpClass()

    def test_VirtualMachine_implements_VMProtocol(self):
        assert isinstance(vm.VirtualMachine, interfaces.VMProtocol)

    def test_verification_fails_for_invalid_proofs(self):
        tree = classes.Tree.from_leaves(self.leaves)
        proof = tree.prove(self.leaves[0])
        proof = vm.decompile(proof)

        wrong_proof = [b'\xffabsdfdfd']
        prover = vm.VirtualMachine(vm.adapt_legacy_proof(wrong_proof))
        assert not prover.run()

        wrong_proof = proof[1:]
        prover = vm.VirtualMachine(wrong_proof)
        assert not prover.run()

        wrong_proof = proof[:-1]
        prover = vm.VirtualMachine(wrong_proof)
        assert not prover.run()

        wrong_proof = [*proof]
        wrong_proof[-1] = wrong_proof[-1] + b'1'
        prover = vm.VirtualMachine(wrong_proof)
        assert not prover.run()

        wrong_proof = [*proof]
        wrong_proof[1] = b'\x99' + wrong_proof[1]
        prover = vm.VirtualMachine(wrong_proof)
        assert not prover.run()

        wrong_proof = [*proof]
        wrong_proof[1] = wrong_proof[1] + b'\x99'
        prover = vm.VirtualMachine(wrong_proof)
        assert not prover.run()

    def test_verification_fails_for_malicious_proofs(self):
        leaves = [b'leaf0', b'leaf1', b'leaf2']
        tree = classes.Tree.from_leaves(leaves)
        legit_proof = tree.prove(b'leaf0')
        legit_proof = vm.decompile(legit_proof)
        # first instruction in legit_proof is a load_left operation
        assert legit_proof[0] is classes.OpCode.load_left_hsize

        # try to trick the validator by inserting malicious leaf then overwriting
        # with the load_left instruction from the legit_proof, then continuing
        # with the legit_proof
        malicious_proof = [
            classes.OpCode.load_left_hsize,
            sha256(b'\x00malicious leaf').digest(),
            *legit_proof
        ]

        prover = vm.VirtualMachine(malicious_proof)
        assert not prover.run()

        bad = sha256(b'bad').digest()
        malicious_proof = [
            *legit_proof,
            classes.OpCode.load_left_hsize,
            bad,
            classes.OpCode.load_right_hsize,
            bad,
            classes.OpCode.hash_final_hsize,
            sha256(b'\x01' + bad + bad).digest()
        ]

        prover = vm.VirtualMachine(malicious_proof)
        assert not prover.run()

    def test_compute_hashes_e2e(self):
        assert len(vm._EMPTY_HASHES) == 0
        vm.compute_empty_hashes()
        assert len(vm._EMPTY_HASHES) == 256

        max_hash = vm._EMPTY_HASHES[-1]

        vm.set_hash_function(lambda p: sha3_256(p).digest())
        vm.compute_empty_hashes()
        assert max_hash != vm._EMPTY_HASHES[-1]

    def test_load_ops_and_hash_final_hszie(self):
        program = bytes(vm.OpCode.load_left) + (2).to_bytes(2, 'big') + b'1234'
        prover = vm.VirtualMachine(program)
        prover.run()
        assert prover.get_register('left') == b'12'

        program = bytes(vm.OpCode.load_right) + (3).to_bytes(2, 'big') + b'1234'
        prover = vm.VirtualMachine(program)
        prover.run()
        assert prover.get_register('right') == b'123'

        program = bytes(vm.OpCode.load_empty_left)
        program += (0).to_bytes(1, 'big')
        program += bytes(vm.OpCode.load_empty_right)
        program += (123).to_bytes(1, 'big')
        prover = vm.VirtualMachine(program)
        prover.run()
        assert prover.registers['left'] == vm._EMPTY_HASHES[0]
        assert prover.registers['right'] == vm._EMPTY_HASHES[123]

        left = vm._EMPTY_HASHES[0]
        right = vm._EMPTY_HASHES[123]
        root = vm.hash_node(left, right)
        program += bytes(vm.OpCode.hash_final_hsize)
        program += root
        prover = vm.VirtualMachine(program)
        assert prover.run()

        prover = vm.VirtualMachine(bytes(vm.OpCode.load_left_hsize) + b'123')
        prover.set_register('size', 2)
        assert not prover.run()
        assert prover.get_register('left') == b'12'

        prover = vm.VirtualMachine(bytes(vm.OpCode.load_right_hsize) + b'123')
        prover.set_register('size', 3)
        assert not prover.run()
        assert prover.get_register('right') == b'123'

    def test_hash_to_level_ops(self):
        leaf = b'123'
        leaf_hash = vm.hash_leaf(leaf)
        program = bytes(vm.OpCode.load_left) + (3).to_bytes(2, 'big') + leaf
        program += bytes(vm.OpCode.hash_to_level_hsize) + b'\x00' + b'\x09' + leaf_hash
        prover = vm.VirtualMachine(program)

        prover.run()
        hash1 = prover.get_register('return')
        assert len(hash1) == 32

        # check against SparseSubTree
        # sst = classes.SparseSubTree(leaf=leaf, level=9)
        # proof = sst.prove()
        # program = vm.compile(*proof)
        # prover = vm.VirtualMachine(program)
        # assert prover.run()
        # hash2 = prover.get_register('return')
        # assert hash1 == hash2, f"{hash1.hex()}\n{hash2.hex()}"

        # another method
        program = bytes(vm.OpCode.load_left) + (3).to_bytes(2, 'big') + leaf
        program += bytes(vm.OpCode.hash_leaf_left) + bytes(vm.OpCode.set_path_auto)
        program += bytes(vm.OpCode.hash_to_level_path) + b'\x00\x09'
        prover = vm.VirtualMachine(program)
        prover.run()
        hash3 = prover.get_register('return')
        assert hash3 == hash1, f"{hash1.hex()}\n{hash3.hex()}"

        # using hash_to_level
        program = bytes(vm.OpCode.load_left) + (3).to_bytes(2, 'big') + leaf
        program += bytes(vm.OpCode.hash_leaf_left)
        program += bytes(vm.OpCode.hash_to_level) + b'\x00\x09' + \
            (32).to_bytes(2, 'big') + leaf_hash
        prover.run()
        hash4 = prover.get_register('return')
        assert hash4 == hash1, f"{hash1.hex()}\n{hash4.hex()}"

    def test_other_hash_ops(self):
        prover = vm.VirtualMachine(bytes(vm.OpCode.hash_left))
        assert not prover.run()
        assert len(prover.get_errors()) == 0
        left = prover.get_register('left')
        assert type(left) is bytes and len(left) == 32

        prover.reset().load_program(bytes(vm.OpCode.hash_right))
        assert not prover.run()
        assert len(prover.get_errors()) == 0
        right = prover.get_register('right')
        assert type(right) is bytes and len(right) == 32

        prover.reset().load_program(bytes(vm.OpCode.hash_final_hsize) + self.root)
        assert not prover.run()
        assert len(prover.get_errors()) == 0
        ret = prover.get_register('return')
        assert type(ret) is bytes and len(ret) == 32

        prover.reset().load_program(bytes(vm.OpCode.hash_mid))
        prover.set_register('left', b'1')
        prover.set_register('right', b'2')
        assert not prover.run()
        assert len(prover.get_errors()) == 0
        ret = prover.get_register('return')
        assert type(ret) is bytes and len(ret) == 32

        prover.reset().load_program(bytes(vm.OpCode.hash_leaf_left))
        assert not prover.run()
        assert len(prover.get_errors()) == 0
        left = prover.get_register('left')
        assert type(left) is bytes and len(left) == 32

        prover.reset().load_program(bytes(vm.OpCode.hash_leaf_right))
        assert not prover.run()
        assert len(prover.get_errors()) == 0
        right = prover.get_register('right')
        assert type(right) is bytes and len(right) == 32

        prover.reset().load_program(bytes(vm.OpCode.hash_leaf_mid))
        assert not prover.run()
        assert len(prover.get_errors()) == 0
        ret = prover.get_register('return')
        assert type(ret) is bytes and len(ret) == 32

        prover.reset().load_program(bytes(vm.OpCode.hash_leaf_bit))
        assert not prover.run()
        assert len(prover.get_errors()) == 0
        assert prover.get_register('bit') is False
        left = prover.get_register('left')
        assert type(left) is bytes and len(left) == 32

        prover.reset().load_program(bytes(vm.OpCode.hash_bit))
        prover.set_register('left', b'1')
        prover.set_register('right', b'2')
        prover.set_register('bit', True)
        assert not prover.run()
        assert len(prover.get_errors()) == 0
        ret = prover.get_register('right')
        assert type(ret) is bytes and len(ret) == 32
        prover.reset().load_program(bytes(vm.OpCode.hash_bit))
        prover.set_register('left', b'1')
        prover.set_register('right', b'2')
        prover.set_register('bit', False)
        assert not prover.run()
        assert len(prover.get_errors()) == 0
        assert ret == prover.get_register('left')

        prover.reset().load_program(bytes(vm.OpCode.hash_final) + b'\x01' + b'x')
        assert not prover.run()
        assert len(prover.get_errors()) == 0
        ret = prover.get_register('return')
        assert type(ret) is bytes and len(ret) == 32

        prover.reset().load_program(bytes(vm.OpCode.hash_with_empty) + b'\x02')
        prover.set_register('left', b'1')
        prover.set_register('bit', False)
        assert not prover.run()
        assert len(prover.get_errors()) == 0, prover.get_errors()
        ret = prover.get_register('left')
        assert type(ret) is bytes and len(ret) == 32

    def test_set_ops(self):
        prover = vm.VirtualMachine(bytes(vm.OpCode.set_hsize) + (69).to_bytes(1, 'big'))
        assert not prover.run()
        assert prover.get_register('size') == 69

        program = bytes(vm.OpCode.set_path) + (2).to_bytes(2, 'big') + b'12'
        prover.reset().load_program(program)
        assert not prover.run()
        assert prover.get_register('path') == b'12'

        program = bytes(vm.OpCode.set_path_hsize) + b'12'
        prover.reset().load_program(program)
        prover.set_register('size', 2)
        assert not prover.run()
        assert prover.get_register('path') == b'12'

        program = bytes(vm.OpCode.set_path_auto)
        prover.reset().load_program(program)
        prover.set_register('left', b'321')
        prover.set_register('right', b'abc')
        assert not prover.run()
        assert prover.get_register('path') == b'321'
        prover.reset().load_program(program)
        prover.set_register('right', b'abc')
        assert not prover.run()
        assert prover.get_register('path') == b'abc'

    def test_other_ops(self):
        prover = vm.VirtualMachine()
        program = bytes(vm.OpCode.get_path_bit) + b'\x01'
        prover.reset().load_program(program)
        prover.set_register('path', (0b01000000).to_bytes(1, 'big'))
        prover.run()
        assert prover.get_register('bit')
        prover.reset().load_program(program)
        prover.set_register('path', (0b10000000).to_bytes(1, 'big'))
        prover.run()
        assert not prover.get_register('bit')

        leaf = b'123'
        leaf = sha256(b'\x00' + leaf).digest()
        root = sha256(b'\x01' + leaf).digest()
        subprogram = bytes(vm.OpCode.load_left) + len(leaf).to_bytes(2, 'big')
        subprogram += leaf + bytes(vm.OpCode.hash_final) + b'\x20' + root

        program = bytes(vm.OpCode.subroutine_left) + b'\x00'
        prover.reset().load_program(program)
        assert not prover.run()
        assert len(prover.get_errors()) > 0
        program = bytes(vm.OpCode.subroutine_left)
        program += len(subprogram).to_bytes(2, 'big') + subprogram
        prover.reset().load_program(program)
        prover.run()
        assert prover.get_register('left')
        assert not prover.get_register('right')
        assert len(prover.get_errors()) == 0

        program = bytes(vm.OpCode.subroutine_right) + b'\x00'
        prover.reset().load_program(program)
        assert not prover.run()
        assert len(prover.get_errors()) > 0
        program = bytes(vm.OpCode.subroutine_right)
        program += len(subprogram).to_bytes(2, 'big') + subprogram
        prover.reset().load_program(program)
        prover.run()
        assert not prover.get_register('left')
        assert prover.get_register('right')
        assert len(prover.get_errors()) == 0

    def test_move_to_ops(self):
        leaf = b'123'
        program = bytes(vm.OpCode.load_left) + (3).to_bytes(2, 'big') + leaf
        program += bytes(vm.OpCode.hash_leaf_left) + bytes(vm.OpCode.set_path_auto)
        program += bytes(vm.OpCode.hash_to_level_path) + b'\x00\x09'
        prover = vm.VirtualMachine(program)
        prover.run()
        hash1 = prover.get_register('return')
        assert hash1 != b''
        assert prover.get_register('left') == b''

        program += bytes(vm.OpCode.move_to_left)
        prover = vm.VirtualMachine(program)
        prover.run()
        hash2 = prover.get_register('left')
        assert prover.get_register('return') == b''
        assert hash2 == hash1

        program = program[:-1] + bytes(vm.OpCode.move_to_right)
        prover = vm.VirtualMachine(program)
        prover.run()
        hash2 = prover.get_register('right')
        assert prover.get_register('return') == b''
        assert hash2 == hash1

    def test_compiler(self):
        # case 1: `op`
        expected = bytes(vm.OpCode.hash_bit)
        observed = vm.compile(vm.OpCode.hash_bit)
        assert observed == expected, f"case 1: {observed} != {expected}"

        # case 2: `op bytes`
        expected = bytes(vm.OpCode.load_left_hsize) + b'123'
        observed = vm.compile(vm.OpCode.load_left_hsize, b'123')
        assert observed == expected, f"case 2: {observed} != {expected}"

        # case 3: `op u8 bytes`
        expected = bytes(vm.OpCode.hash_final) + b'\x03' + b'123'
        observed = vm.compile(vm.OpCode.hash_final, b'123')
        assert observed == expected, f"case 3: {observed} != {expected}"

        # case 4: `op u16 bytes`
        expected = bytes(vm.OpCode.subroutine_left) + (4).to_bytes(2, 'big') + b'abcd'
        observed = vm.compile(vm.OpCode.subroutine_left, b'abcd')
        assert observed == expected, f"case 4: {observed} != {expected}"

        # case 5: `op u8 u8 u16 bytes`
        expected = bytes(vm.OpCode.hash_to_level) + b'\x00\x09' + (4).to_bytes(2, 'big') + b'abcd'
        observed = vm.compile(vm.OpCode.hash_to_level, 0, 9, b'abcd')
        assert observed == expected, f"case 5: {observed} != {expected}"

        # case 6: `op u8 u8 bytes`
        expected = bytes(vm.OpCode.hash_to_level_hsize) + b'\x00\x09' + b'abcd'
        observed = vm.compile(vm.OpCode.hash_to_level_hsize, 0, 9, b'abcd')
        assert observed == expected, f"case 6: {observed} != {expected}"

        # case 7: `op u8`
        expected = bytes(vm.OpCode.set_hsize) + b'\x08'
        observed = vm.compile(vm.OpCode.set_hsize, 8)
        assert observed == expected, f"case 7: {observed} != {expected}"

        # case 8: `op u8 u8`
        expected = bytes(vm.OpCode.hash_to_level_path) + b'\x01\x09'
        observed = vm.compile(vm.OpCode.hash_to_level_path, 1, 9)
        assert observed == expected, f"case 8: {observed} != {expected}"

    def test_compiler_syntax_errors(self):
        # case 2: `op bytes` required but supply `op`
        with self.assertRaises(SyntaxError) as e:
            vm.compile(vm.OpCode.load_left_hsize)
        assert 'expected param bytes' in str(e.exception)

        # case 2: `op bytes` required but supply `op int`
        with self.assertRaises(SyntaxError) as e:
            vm.compile(vm.OpCode.load_left_hsize, 18)
        assert 'expected bytes' in str(e.exception)

        # case 2: `op bytes(len<=65535)` required but supply `op bytes(len=65536)`
        with self.assertRaises(SyntaxError) as e:
            vm.compile(vm.OpCode.load_left_hsize, b''.join([b'1' for _ in range(65536)]))
        assert 'too large' in str(e.exception)

        # case 3: `op u8 bytes` but supply `op int`
        with self.assertRaises(SyntaxError) as e:
            vm.compile(vm.OpCode.hash_final, 69)
        assert 'expected bytes' in str(e.exception)

        # case 3: `op u8 bytes(len<=255)` required but supply `op bytes(len=256)`
        with self.assertRaises(SyntaxError) as e:
            vm.compile(vm.OpCode.hash_final, b''.join([b'1' for _ in range(256)]))
        assert 'too large' in str(e.exception)

        # case 4: `op u16 bytes(len<=65535)` required but supply `op int`
        with self.assertRaises(SyntaxError) as e:
            vm.compile(vm.OpCode.load_left, 69)
        assert 'expected bytes' in str(e.exception)

        # case 4: `op u16 bytes(len<=65535)` required but supply `op bytes(len=65536)`
        with self.assertRaises(SyntaxError) as e:
            vm.compile(vm.OpCode.load_left, b''.join([b'1' for _ in range(65536)]))
        assert 'too large' in str(e.exception)

        # case 5: `op u8 u8 u16 bytes` required but supply `op u8`
        with self.assertRaises(SyntaxError) as e:
            vm.compile(vm.OpCode.hash_to_level, 22)
        assert "expected params" in str(e.exception)

        # case 5: `op u8 u8 u16 bytes` required but supply `op u16 u8 bytes`
        with self.assertRaises(SyntaxError) as e:
            vm.compile(vm.OpCode.hash_to_level, 2222, 25, b'asd')
        assert "<=255" in str(e.exception)

        # case 6: `op u8 u8 bytes` required but supply `op u8 u8 u8`
        with self.assertRaises(SyntaxError) as e:
            vm.compile(vm.OpCode.hash_to_level_hsize, 8, 8, 8)
        assert "expected bytes" in str(e.exception)

        # case 6: `op u8 u8 bytes` required but supply `op bytes u8 u8`
        with self.assertRaises(SyntaxError) as e:
            vm.compile(vm.OpCode.hash_to_level_hsize, b'abdf', 8, 8)
        assert "expected int" in str(e.exception)

        # case 7: `op u8` required but supply `op bytes`
        with self.assertRaises(SyntaxError) as e:
            vm.compile(vm.OpCode.set_hsize, b'abdf')
        assert "expected int" in str(e.exception)

        # case 8: `op u8 u8` required but supply `op bytes u8`
        with self.assertRaises(SyntaxError) as e:
            vm.compile(vm.OpCode.hash_to_level_path, b'abdf', 8)
        assert "expected int" in str(e.exception)

    def test_compile_decompile_e2e(self):
        src = [
            vm.OpCode.set_hsize, 3,
            vm.OpCode.load_left_hsize, b'123',
            vm.OpCode.load_left, b'123',
            vm.OpCode.hash_to_level_hsize, 0, 9, b'abc',
            vm.OpCode.hash_to_level_path, 1, 9,
            vm.OpCode.hash_final, b'321'
        ]
        compiled = vm.compile(*src)
        observed = vm.decompile(compiled)
        assert observed == src, f"{observed}\n{src}"

        src = [
            vm.OpCode.load_left, b'321',
            vm.OpCode.load_right, b'abc',
            vm.OpCode.hash_xor_left,
            vm.OpCode.hash_xor_right,
            vm.OpCode.hash_final, b'321'
        ]
        compiled = vm.compile(*src)
        observed = vm.decompile(compiled)
        assert observed == src, f"{observed}\n{src}"


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from .errors import tert
from .interfaces import PackableProtocol
from typing import Any
import struct


def serialize_part(data: Any) -> bytes:
    """Serializes an instance of a PackableProtocol implementation or
        built-in type, recursively calling itself as necessary.
    """
    tert(isinstance(data, PackableProtocol) or \
        type(data) in (list, set, tuple, str, bytes, bytearray, int, float) or \
            data is None,
        'data type must be one of (PackableProtocol, list, set, tuple, ' + \
        'str, bytes, bytearray, int, float, NoneType)')

    if isinstance(data, PackableProtocol):
        packed = bytes(data.__class__.__name__, 'utf-8').hex()
        packed = bytes(packed, 'utf-8') + b'_' + data.pack()
        return struct.pack(
            f'!1sI{len(packed)}s',
            b'p',
            len(packed),
            packed
        )

    if type(data) in (list, set, tuple):
        items = b''.join([serialize_part(item) for item in data])
        code = ({
            list: b'l',
            set: b'e',
            tuple: b't'
        })[type(data)]

        return struct.pack(
            f'!1sI{len(items)}s',
            code,
            len(items),
            items
        )

    if type(data) in (bytes, bytearray):
        return struct.pack(
            f'!1sI{len(data)}s',
            b'b' if type(data) is bytes else b'a',
            len(data),
            data
        )

    if type(data) is str:
        data = bytes(data, 'utf-8')
        return struct.pack(
            f'!1sI{len(data)}s',
            b's',
            len(data),
            data
        )

    if type(data) is int:
        return struct.pack(
            f'!1sII',
            b'i',
            4,
            data
        )

    if type(data) is float:
        return struct.pack(
            f'!1sId',
            b'f',
            8,
            data
        )

    if data is None:
        return struct.pack(
            f'!1sI',
            b'n',
            0
        )


def deserialize_part(data: bytes, inject: dict = {}) -> Any:
    """Deserializes an instance of a PackableProtocol implementation
        or built-in type, recursively calling itself as necessary.
    """
    code, data = struct.unpack(f'!1s{len(data)-1}s', data)
    dependencies = {**globals(), **inject}

    if code == b'p':
        packed_len, data = struct.unpack(f'!I{len(data)-4}s', data)
        packed, _ = struct.unpack(f'!{packed_len}s{len(data)-packed_len}s', data)
        packed_class, _, packed_data = packed.partition(b'_')
        packed_class = str(bytes.fromhex(str(packed_class, 'utf-8')), 'utf-8')
        tert(packed_class in dependencies,
            f'{packed_class} not found in globals or inject; cannot unpack')
        tert(hasattr(dependencies[packed_class], 'unpack'),
            f'{packed_class} must have unpack method')
        return dependencies[packed_class].unpack(packed_data, inject=inject)

    if code in (b'l', b'e', b't'):
        let_len, data = struct.unpack(f'!I{len(data)-4}s', data)
        let_data, _ = struct.unpack(f'!{let_len}s{len(data)-let_len}s', data)
        items = []
        while len(let_data) > 0:
            _, item_len, _ = struct.unpack(f'!1sI{len(let_data)-5}s', let_data)
            item, let_data = struct.unpack(
                f'!{5+item_len}s{len(let_data)-5-item_len}s',
                let_data
            )
            items.append(deserialize_part(item, inject=inject))

        if code == b'l':
            return items
        if code == b'e':
            return set(items)
        if code == b't':
            return tuple(items)

    if code in (b'b', b'a'):
        bt_len, data = struct.unpack(f'!I{len(data)-4}s', data)
        bt_data, _ = struct.unpack(f'!{bt_len}s{len(data)-bt_len}s', data)
        return bt_data if code == b'b' else bytearray(bt_data)

    if code == b's':
        s_len, data = struct.unpack(f'!I{len(data)-4}s', data)
        s, _ = struct.unpack(f'!{s_len}s{len(data)-s_len}s', data)
        return str(s, 'utf-8')

    if code == b'i':
        return struct.unpack(f'!II{len(data)-8}s', data)[1]

    if code == b'f':
        return struct.unpack(f'!Id{len(data)-12}s', data)[1]

    if code == b'n':
        return None



"""
Merkleasy is a Python library for Merkle Trees. It consists of a virtual machine,
a Tree class, an XorHashTree class, and a set of miscellaneous functions and
other classes.

Version 0.1.2.
"""

from .classes import Tree, set_hash_function, get_hash_function
from .errors import ImplementationError, SecurityError
from .vm import VirtualMachine, OpCode, compile, decompile, hash_node, hash_leaf
from .xorhashtree import XorHashTree


__version__ = '0.1.2'



from __future__ import annotations
from .errors import tert, eruces, vert, SecurityError
from .vm import (
    get_hash_function,
    VirtualMachine,
    OpCode,
    compile,
    decompile,
    hash_leaf,
    xor
)
from typing import Optional
import json


class XorHashTree:
    """An XOR Hash Tree. It is a node with a left and right branch, and the
        root is the XOR of the hashes of the left and right branches.
    """
    root: bytes
    parent: Optional[XorHashTree]
    left: XorHashTree | bytes
    right: XorHashTree | bytes
    left_bytes: bytes
    right_bytes: bytes

    @property
    def root(self) -> bytes:
        return self._root

    @property
    def parent(self) -> Optional[XorHashTree]:
        return self._parent

    @parent.setter
    def parent(self, parent) -> None:
        """Set the parent. Raises `TypeError` upon invalid input."""
        tert(isinstance(parent, XorHashTree) or parent is None, 'invalid parent')
        self._parent = parent

    @property
    def left(self) -> XorHashTree | bytes:
        return self._left

    @property
    def right(self) -> XorHashTree | bytes:
        return self._right

    @property
    def left_bytes(self) -> XorHashTree | bytes:
        return self._left_bytes

    @property
    def right_bytes(self) -> XorHashTree | bytes:
        return self._right_bytes

    def __init__(self, left: XorHashTree | bytes, right: XorHashTree | bytes) -> None:
        """Set the left, right, and calculated root."""
        tert(type(left) in (XorHashTree, bytes, bytearray),
             'left must be one of XorHashTree, bytes, bytearray')
        tert(type(right) in (XorHashTree, bytes, bytearray),
             'right must be one of XorHashTree, bytes, bytearray')

        self._left = left if type(left) in (XorHashTree, bytes) else bytes(left)
        self._right = right if type(right) in (XorHashTree, bytes) else bytes(right)
        self.parent = None
        _hsize = len(hash_leaf(b''))

        if type(self._left) is bytes and len(self._left) != _hsize:
            self._left = hash_leaf(self._left)
        if type(self._right) is bytes and len(self._right) != _hsize:
            self._right = hash_leaf(self._right)

        if type(self.left) is XorHashTree:
            self.left.parent = self
        if type(self.right) is XorHashTree:
            self.right.parent = self

        self._left_bytes = self.left.root if isinstance(self.left, XorHashTree) else self.left
        self._right_bytes = self.right.root if isinstance(self.right, XorHashTree) else self.right
        self._root = xor(get_hash_function()(self.left_bytes), get_hash_function()(self.right_bytes))

    def __str__(self) -> str:
        """Return the root, left, and right in hexadecimal."""
        return f'({self.root.hex()} [{self.left_bytes.hex()}, {self.right_bytes.hex()}])'

    def __repr__(self) -> str:
        """Return the root, left, and right in hexadecimal recursively."""
        left = self.left.hex() if isinstance(self.left, bytes) else repr(self.left)
        right = self.right.hex() if isinstance(self.right, bytes) else repr(self.right)
        return f'({self.root.hex()} [{left}, {right}])'

    def __eq__(self, other: object) -> bool:
        return type(other) is XorHashTree and self.root == other.root and \
            self.left == other.left and self.right == other.right

    def __hash__(self) -> int:
        return hash(repr(self))

    def to_dict(self) -> dict:
        """Serialize to a dict."""
        left = self.left.to_dict() if type(self.left) is XorHashTree else self.left.hex()
        right = self.right.to_dict() if type(self.right) is XorHashTree else self.right.hex()

        return {
            self.root.hex(): [left, right]
        }

    def to_json(self, pretty: bool = False) -> str:
        """Serialize to json."""
        opts = {} if not pretty else {'indent': '\t'}
        return json.dumps(self.to_dict(), **opts)

    @classmethod
    def from_leaves(cls, leaves: list[bytes]) -> XorHashTree:
        """Return a full tree constructed from the leaves. Raises
            `TypeError` or `ValueError` upon invalid input.
        """
        tert(type(leaves) in (tuple, list), 'leaves must be tuple or list of bytes')
        vert(len(leaves) >= 2, 'must have at least 2 leaves')

        for leaf in leaves:
            tert(isinstance(leaf, bytes), 'leaves must be tuple or list of bytes')

        # hash all leaves
        parts = [hash_leaf(leaf) for leaf in leaves]

        # recursively join until reaching the root
        while len(parts) > 1:
            parts = _join(parts)

        return parts[0]

    @classmethod
    def from_dict(cls, data: dict) -> XorHashTree:
        """Deserialize from a dict and return an instance. Raises
            `TypeError`, `ValueError`, or `SecurityError` if the dict
            does not encode a valid `XorHashTree` instance.
        """
        tert(type(data) is dict, 'data must be dict type')
        vert(len(data.keys()) == 1, 'data must have one key')
        root = list(data.keys())[0]
        vert(len(data[root]) == 2, 'data[root] must have left and right branch')
        left = data[root][0]
        right = data[root][1]

        left = bytes.fromhex(left) if type(left) is str else cls.from_dict(left)
        right = bytes.fromhex(right) if type(right) is str else cls.from_dict(right)
        tree = cls(left, right)
        eruces(tree.root.hex() == root, 'root mismatch')
        return tree

    @classmethod
    def from_json(cls, data: str) -> XorHashTree:
        """Deserialize from json and return an instance. Raises
            `json.decoder.JSONDecodeError` upon invalid input. Raises
            `TypeError`, `ValueError`, or `SecurityError` if the JSON
            does not encode a valid `XorHashTree` instance.
        """
        return cls.from_dict(json.loads(data))

    def prove(self, leaf: bytes, verbose: bool = False) -> bytes:
        """Create an inclusion proof for a leaf. Use verbose=True to add
            hash checks at each tree level. Raises `TypeError` or
            `ValueError` upon invalid input. Return value consists of
            `OpCode`s and their arguments compiled into bytes.
        """
        tert(type(leaf) is bytes, 'leaf must be bytes')
        leaf_hash = hash_leaf(leaf)

        # get set of nodes
        nodes = set(_traverse(self, tuple(), False))
        node_hashes = [n[0] for n in nodes]

        vert(leaf in node_hashes or leaf_hash in node_hashes,
            'the given leaf was not found in the tree')

        # start at the leaf node
        node = [n for n in nodes if n[0] in (leaf, leaf_hash)][0]

        # reverse the history
        history = list(node)[2][::-1]
        proof = []
        first = True

        for direction in history:
            step = [OpCode.hash_xor_left] if not first else []
            if first or verbose:
                step.extend([OpCode.load_left_hsize, node[0]])
                first = False
            if direction == -1:
                step.extend([OpCode.load_right_hsize, node[1].right_bytes])
            else:
                step.extend([OpCode.load_right_hsize, node[1].left_bytes])

            proof.append(tuple(step))

            # advance
            if node[1] is not self:
                new_node = node[1]
                node = [n for n in nodes if n[0] == new_node.root][0]

        proof.append((OpCode.hash_xor_final, self.root))

        if len(leaf_hash) != 32:
            proof = [OpCode.set_hsize, len(leaf_hash), *proof]

        return compile(*proof)

    @staticmethod
    def verify(root: bytes, leaf: bytes, proof: bytes|list[tuple[OpCode|bytes,]],
               report_errors: bool = False
    ) -> bool|tuple[bool, list[BaseException,]]:
        """Verify an inclusion proof is valid. If report_errors is True,
            returns status and errors. Otherwise, returns status. Can
            accept compiled or decompiled proofs.
        """
        tert(type(root) is bytes, 'root must be bytes')
        tert(type(leaf) is bytes, 'leaf must be bytes')
        tert(type(proof) in (list, bytes),
             'proof must be bytes or list of tuple[OpCode|bytes,]')

        if type(proof) is list:
            tert(all(isinstance(p, tuple) for p in proof)
                 or all(type(p) in (bytes, OpCode) for p in proof),
                 'proof must be list of tuple[OpCode|bytes,]')
            if type(proof[0]) is tuple:
                tert(all(type(c) in (OpCode, bytes) for step in proof for c in step),
                     'proof must be list of tuple[OpCode|bytes,]')
            try:
                proof = compile(*proof)
            except Exception as e:
                if report_errors:
                    return (False, (e,))
                return False

        try:
            decompiled = decompile(proof)
        except Exception as e:
            if report_errors:
                return (False, (e,))
            return False

        leaf_hash = hash_leaf(leaf)
        index = 3 if decompiled[0] is OpCode.set_hsize else 1
        if decompiled[index] not in (leaf, leaf_hash):
            if report_errors:
                return (False, (SecurityError('proof does not reference leaf'),))
            return False

        if root not in proof:
            if report_errors:
                return (False, (SecurityError('proof does not reference root'),))
            return False

        vm = VirtualMachine(proof)
        if vm.run():
            return (True, []) if report_errors else True
        return (False, vm.get_errors()) if report_errors else False


def _join(parts: list[bytes]) -> list[XorHashTree]:
    """Joins every two items together, returning the resulting list of XorHashTrees."""
    new_parts = []

    # join every two together
    for i in range(0, len(parts), 2):
        if i+1 < len(parts):
            new_parts.append(XorHashTree(parts[i], parts[i+1]))
        else:
            new_parts.append(parts[i])

    return new_parts

def _traverse(branch: XorHashTree, history: tuple[int], exclude_root: bool = True
              ) -> list[tuple[XorHashTree, XorHashTree|None, tuple[int,]]]:
    """Returns form [(hash, parent, history),...]."""
    nodes = []
    if not exclude_root:
        root = (branch.root, None, tuple(history))
        nodes += [root]
    left = (branch.left_bytes, branch, (*history, -1))
    right = (branch.right_bytes, branch, (*history, 1))
    nodes += [left, right]

    if type(branch.left) is XorHashTree:
        nodes.extend(_traverse(branch.left, (*history, -1)))
    if type(branch.right) is XorHashTree:
        nodes.extend(_traverse(branch.right, (*history, 1)))

    return nodes



from context import errors, xorhashtree, vm
from hashlib import sha256, shake_256
from random import randint
import unittest


class TestXorHashTree(unittest.TestCase):
    """Test suite for the XorHashTree class."""
    def setUp(self) -> None:
        self.original_hash_function = vm.get_hash_function()

    def tearDown(self) -> None:
        vm.set_hash_function(self.original_hash_function)

    def hash(self, data: bytes) -> bytes:
        return sha256(data).digest()

    @staticmethod
    def print_proof(proof: list[tuple[vm.OpCode|bytes,]]):
        formatted_proof = []
        for step in proof:
            formatted_proof.append(tuple([
                c.name if type(c) is vm.OpCode else c.hex()
                for c in step
            ]))
        print(formatted_proof)

    def test_xorhashtree_has_class_XorHashTree(self):
        assert hasattr(xorhashtree, 'XorHashTree')
        assert isinstance(xorhashtree.XorHashTree, type)

    def test_XorHashTree_joins_left_and_right_into_root(self):
        left = self.hash(b'\x00hello')
        right = self.hash(b'\x00world')
        joined = vm.xor(self.hash(left), self.hash(right))
        tree = xorhashtree.XorHashTree(left, right)

        assert hasattr(tree, 'root')
        assert type(tree.root) is bytes
        assert tree.root == joined

    def test_XorHashTree_from_leaves_hashes_and_joins_leaves(self):
        leaves = [b'hello', b'world']
        left_hash = self.hash(b'\x00' + leaves[0])
        right_hash = self.hash(b'\x00' + leaves[1])
        root = vm.xor(self.hash(left_hash), self.hash(right_hash))

        assert hasattr(xorhashtree.XorHashTree, 'from_leaves')
        tree = xorhashtree.XorHashTree.from_leaves(leaves)
        assert tree.left == left_hash
        assert tree.right == right_hash
        assert tree.root == root

    def test_XorHashTree_from_leaves_requires_at_least_two_leaves(self):
        with self.assertRaises(ValueError) as e:
            xorhashtree.XorHashTree.from_leaves([b'123'])
        assert str(e.exception) == 'must have at least 2 leaves'

    def test_XorHashTree_from_leaves_joins_any_number_of_leaves(self):
        roots = set()

        for i in range(2, 300):
            leaves = [n.to_bytes(2, 'big') for n in range(i)]
            tree = xorhashtree.XorHashTree.from_leaves(leaves)
            assert tree.root not in roots
            roots.add(tree.root)

    def test_XorHashTree_instance_serializes_to_dict(self):
        tree = xorhashtree.XorHashTree(b'left', b'right')
        assert hasattr(tree, 'to_dict') and callable(tree.to_dict)
        serialized = tree.to_dict()
        assert type(serialized) is dict

    def test_XorHashTree_from_dict_deserializes_and_returns_instance(self):
        tree = xorhashtree.XorHashTree(b'left', b'right')
        serialized = tree.to_dict()
        assert hasattr(xorhashtree.XorHashTree, 'from_dict')
        deserialized = xorhashtree.XorHashTree.from_dict(serialized)
        assert type(deserialized) is xorhashtree.XorHashTree
        assert tree == deserialized

    def test_XorHashTree_from_dict_raises_errors_for_invalid_params(self):
        tree = xorhashtree.XorHashTree(b'left', b'right')
        serialized = tree.to_dict()

        with self.assertRaises(TypeError) as e:
            xorhashtree.XorHashTree.from_dict('not a dict')
        assert str(e.exception) == 'data must be dict type'

        with self.assertRaises(ValueError) as e:
            xorhashtree.XorHashTree.from_dict({})
        assert str(e.exception) == 'data must have one key'

        with self.assertRaises(ValueError) as e:
            xorhashtree.XorHashTree.from_dict({**serialized, 'what': 'huh'})
        assert str(e.exception) == 'data must have one key'

        with self.assertRaises(ValueError) as e:
            xorhashtree.XorHashTree.from_dict({"3213": [1,2,3]})
        assert str(e.exception) == 'data[root] must have left and right branch'

        with self.assertRaises(errors.SecurityError) as e:
            key = list(serialized.keys())[0]
            xorhashtree.XorHashTree.from_dict({"2323": serialized[key]})
        assert str(e.exception) == 'root mismatch'

    def test_XorHashTree_instance_serializes_to_json(self):
        tree = xorhashtree.XorHashTree(b'left', b'right')
        assert hasattr(tree, 'to_json') and callable(tree.to_json)
        serialized = tree.to_json()
        assert type(serialized) is str

    def test_XorHashTree_from_json_unserializes_and_returns_instance(self):
        tree = xorhashtree.XorHashTree(b'left', b'right')
        serialized = tree.to_json()
        assert hasattr(xorhashtree.XorHashTree, 'from_json')
        deserialized = xorhashtree.XorHashTree.from_json(serialized)
        assert type(deserialized) is xorhashtree.XorHashTree
        assert tree == deserialized

    def test_XorHashTree_prove_produces_bytes_proof(self):
        for i in range(2, 300):
            leaves = [n.to_bytes(2, 'big') for n in range(i)]
            tree = xorhashtree.XorHashTree.from_leaves(leaves)
            leaf = randint(0, i-1).to_bytes(2, 'big')
            proof = tree.prove(leaf)
            proof_verbose = tree.prove(leaf, verbose=True)
            assert type(proof) is bytes
            assert type(proof_verbose) is bytes
            decompiled = vm.decompile(proof)
            decompiled_verbose = vm.decompile(proof_verbose)
            assert len(decompiled_verbose) >= len(decompiled)

    def test_XorHashTree_prove_raises_errors_for_invalid_params(self):
        leaves = [n.to_bytes(2, 'big') for n in range(13)]
        tree = xorhashtree.XorHashTree.from_leaves(leaves)

        with self.assertRaises(TypeError) as e:
            tree.prove('not bytes')
        assert str(e.exception) == 'leaf must be bytes'

        with self.assertRaises(ValueError) as e:
            tree.prove(b'not in tree')
        assert str(e.exception) == 'the given leaf was not found in the tree'

    def test_XorHashTree_prove_result_decompiles(self):
        leaves = [n.to_bytes(2, 'big') for n in range(17)]
        tree = xorhashtree.XorHashTree.from_leaves(leaves)
        leaf_to_prove = randint(3, 15).to_bytes(2, 'big')
        proof = tree.prove(leaf_to_prove)
        assert type(proof) is bytes
        decompiled = vm.decompile(proof)
        assert type(decompiled) is list

    def test_verbose_proof_is_longer_and_has_2_load_steps_after_hash(self):
        leaves = [n.to_bytes(2, 'big') for n in range(17)]
        tree = xorhashtree.XorHashTree.from_leaves(leaves)
        leaf_to_prove = randint(3, 15).to_bytes(2, 'big')
        proof = tree.prove(leaf_to_prove)
        # self.print_proof(proof)
        verbose = tree.prove(leaf_to_prove, verbose=True)

        assert len(verbose) > len(proof)

        # normal proof
        proof = vm.decompile(proof)
        assert vm.OpCode.load_left_hsize in proof[:4], proof[:4]
        assert vm.OpCode.load_right_hsize in proof[:4], proof[:4]
        for i in range(4, len(proof)-2, 3):
            step = proof[i]
            assert step is vm.OpCode.hash_xor_left, (i, step)
        assert proof[-2] is vm.OpCode.hash_xor_final

        # verbose proof
        verbose = vm.decompile(verbose)
        for i in range(0, len(verbose)-2, 5):
            step = verbose[i]
            assert step is vm.OpCode.load_left_hsize, (i, step)
        assert verbose[-2] is vm.OpCode.hash_xor_final

    def test_XorHashTree_verify_executes_without_error_for_valid_proof(self):
        for i in range(2, 300):
            leaves = [n.to_bytes(2, 'big') for n in range(i)]
            tree = xorhashtree.XorHashTree.from_leaves(leaves)
            leaf = randint(0, i-1).to_bytes(2, 'big')
            proof = tree.prove(leaf)
            xorhashtree.XorHashTree.verify(tree.root, leaf, proof)
            xorhashtree.XorHashTree.verify(tree.root, leaf, vm.decompile(proof))

    def test_XorHashTree_verify_raises_errors_for_invalid_params(self):
        leaves = [n.to_bytes(2, 'big') for n in range(13)]
        tree = xorhashtree.XorHashTree.from_leaves(leaves)
        leaf = leaves[3]
        proof = tree.prove(leaf)

        with self.assertRaises(TypeError) as e:
            xorhashtree.XorHashTree.verify('tree.root', leaf, proof)
        assert str(e.exception) == 'root must be bytes'

        with self.assertRaises(TypeError) as e:
            xorhashtree.XorHashTree.verify(tree.root, 'leaf', proof)
        assert str(e.exception) == 'leaf must be bytes'

        with self.assertRaises(TypeError) as e:
            xorhashtree.XorHashTree.verify(tree.root, leaf, {'not': 'list'})
        assert str(e.exception) == 'proof must be bytes or list of tuple[OpCode|bytes,]'

        with self.assertRaises(TypeError) as e:
            wrong_proof = ['not bytes']
            xorhashtree.XorHashTree.verify(tree.root, leaf, wrong_proof)
        assert str(e.exception) == 'proof must be list of tuple[OpCode|bytes,]', str(e.exception)

    def test_XorHashTree_verify_returns_False_and_errors_for_invalid_proofs(self):
        leaves = [n.to_bytes(3, 'big') for n in range(13)]
        tree = xorhashtree.XorHashTree.from_leaves(leaves)
        leaf = leaves[3]
        proof = tree.prove(leaf)
        proof = vm.decompile(proof)

        result = xorhashtree.XorHashTree.verify(tree.root, leaf, proof, True)
        assert result[0], result[1]
        assert len(result[1]) == 0, result[1]

        # test proof with wrong leaf
        result = xorhashtree.XorHashTree.verify(tree.root, b'\x00', proof, True)
        assert not result[0], result[1]
        assert len(result[1]) > 0, result[1]
        assert type(result[1][0]) is errors.SecurityError
        assert str(result[1][0]) == 'proof does not reference leaf'

        wrong_proof = proof[2:]
        result = xorhashtree.XorHashTree.verify(tree.root, leaf, wrong_proof, True)
        assert len(result[1]) == 1
        assert type(result[1][0]) is errors.SecurityError
        assert str(result[1][0]) == 'proof does not reference leaf'

        wrong_proof = proof[:-1]
        assert not xorhashtree.XorHashTree.verify(tree.root, leaf, wrong_proof)

        wrong_proof = [*proof]
        wrong_proof[-1] = wrong_proof[-1][:-4] + b'\xfe\xed\xbe\xef'
        assert not xorhashtree.XorHashTree.verify(tree.root, leaf, wrong_proof)

        wrong_proof = [*proof]
        wrong_proof = [*wrong_proof[:2], b'\x90', *wrong_proof[2:]]
        result = xorhashtree.XorHashTree.verify(tree.root, leaf, wrong_proof, True)
        assert not result[0]
        assert "expected OpCode" in str(result[1][0]), str(result[1][0])

        wrong_proof = [*proof]
        wrong_proof[1] = wrong_proof[1][:-1] + b'\x99'
        assert not xorhashtree.XorHashTree.verify(tree.root, leaf, wrong_proof)

    def test_XorHashTree_verify_does_not_validate_malicious_proof(self):
        leaves = [b'leaf0', b'leaf1', b'leaf2']
        tree = xorhashtree.XorHashTree.from_leaves(leaves)
        legit_proof = tree.prove(b'leaf0')
        legit_proof = vm.decompile(legit_proof)

        # first instruction in legit_proof is a load_left operation
        assert legit_proof[0] is vm.OpCode.load_left_hsize

        # try to trick the validator by inserting malicious leaf then overwriting
        # with the load_left_hsize instruction from the legit_proof, then
        # continuing with the legit_proof
        malicious_proof = [
            vm.OpCode.load_left_hsize, sha256(b'malicious leaf').digest(),
            *legit_proof
        ]

        # prevent proof hijacking
        assert not xorhashtree.XorHashTree.verify(tree.root, b'malicious leaf', malicious_proof)

        # try to trick the validator by using a proof for a different tree
        malicious_proof = xorhashtree.XorHashTree.from_leaves([b'malicious', b'leaves']).prove(b'malicious')
        result = xorhashtree.XorHashTree.verify(tree.root, b'malicious', malicious_proof, True)
        assert not result[0]
        assert type(result[1][0]) is errors.SecurityError
        assert str(result[1][0]) == 'proof does not reference root'

    def test_e2e_arbitrary_branching(self):
        leaves = [sha256(n.to_bytes(2, 'big')).digest() for n in range(13)]

        tree = xorhashtree.XorHashTree(leaves[0], leaves[1])
        for i in range(2, len(leaves)):
            if randint(0, 1) == 0:
                tree = xorhashtree.XorHashTree(tree, leaves[i])
            else:
                tree = xorhashtree.XorHashTree(leaves[i], tree)

        leaf = leaves[randint(0, len(leaves)-1)]
        proof1 = tree.prove(leaf)
        proof2 = tree.prove(leaf, verbose=True)
        xorhashtree.XorHashTree.verify(tree.root, leaf, proof1)
        xorhashtree.XorHashTree.verify(tree.root, leaf, proof2)

    def test_e2e_alternate_hash_size(self):
        vm.set_hash_function(lambda preimage: shake_256(preimage).digest(24))
        leaves = [n.to_bytes(2, 'big') for n in range(13)]
        tree = xorhashtree.XorHashTree.from_leaves(leaves)
        leaf = leaves[randint(0, len(leaves)-1)]
        proof = tree.prove(leaf)
        assert xorhashtree.XorHashTree.verify(tree.root, leaf, proof)

    def test_prove_all_mirror_XorHashTrees_share_same_root(self):
        """Potential security issue that must be noted: all trees that
            have identical right and left branches will have the same
            root, regardless of what those branches are. It is crucial
            that mirrored XorHashTrees are never used (except perhaps as
            a CTF challenge).
        """
        leaves = [b'leaf0', b'leaf1', b'leaf2']
        tree0 = xorhashtree.XorHashTree(leaves[0], leaves[0])
        tree1 = xorhashtree.XorHashTree(leaves[1], leaves[1])
        tree2 = xorhashtree.XorHashTree(leaves[2], leaves[2])

        assert tree0.root == tree1.root == tree2.root


if __name__ == '__main__':
    unittest.main()



from context import sparse, vm
import unittest


class TestSparseSubTree(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.leaves = [i.to_bytes(1, 'big') for i in range(10)]
        return super().setUpClass()

    def test_calculate_intersection(self):
        idx1 = 0b0100
        idx2 = 0b0110
        point = sparse.calculate_intersection(idx1, idx2)
        assert point == 2

        idx1 = 0b101000
        idx2 = 0b010110
        point = sparse.calculate_intersection(idx1, idx2)
        assert point == 6

        idx1 = 0b010100
        idx2 = 0b010100
        point = sparse.calculate_intersection(idx1, idx2)
        assert point == 0

    def test_intersection_level(self):
        st1 = sparse.SparseSubTree(b'test1', 255)
        st2 = sparse.SparseSubTree(b'test2', 255)
        level = st1.intersection_level(st2)
        assert type(level) is int
        assert -1 <= level <= 255
        bm1 = st1.leaf_hash
        bm2 = st2.leaf_hash
        assert level == sparse.calculate_intersection(bm1, bm2)

    def test_serialization_e2e(self):
        st1 = sparse.SparseSubTree(b'test1', 255)
        st2 = sparse.SparseSubTree(b'test2', 255)
        packed1 = st1.pack()
        packed2 = st2.pack()
        assert type(packed1) == bytes
        assert type(packed2) == bytes
        unpacked1 = sparse.SparseSubTree.unpack(packed1)
        unpacked2 = sparse.SparseSubTree.unpack(packed2)
        assert type(unpacked1) is sparse.SparseSubTree
        assert type(unpacked2) is sparse.SparseSubTree
        assert unpacked1 == st1
        assert unpacked2 == st2

    def test_proof_returns_list_of_compilable_tuples(self):
        st1 = sparse.SparseSubTree(leaf=b'123', level=2)
        proof = st1.prove()
        assert type(proof) is list
        for step in proof:
            assert type(step) is tuple
            for item in step:
                assert type(item) in [bytes, vm.OpCode, int]
        assert len(proof) == 2*st1.level + 3

        # important for saving bytecode size across whole proof
        assert proof[0][0] is vm.OpCode.set_hsize
        assert proof[-1][0] is vm.OpCode.hash_final_hsize

        program = vm.compile(*proof)
        assert len(program) > 0 and type(program) is bytes

    def test_proof_validates(self):
        st1 = sparse.SparseSubTree(leaf=b'123', level=2)
        proof = st1.prove()
        program = vm.compile(*proof)
        prover = vm.VirtualMachine(program)
        assert prover.run()
        assert prover.registers['return'] is not None
        assert prover.registers['return'] == proof[-1][1], \
            f"{prover.registers['return'].hex()}\n{proof[-1][1].hex()}"

    def test_path_returns_list_of_bytes(self):
       sst = sparse.SparseSubTree(leaf=b'123', level=9)
       path = sst.path()
       assert type(path) is list
       assert all([type(item) is bytes for item in path])
       program = b''.join(path)
       prover = vm.VirtualMachine(program)
       assert not prover.run()
       assert prover.has_completed()
       assert len(prover.get_errors()) == 0, prover.get_errors()


class TestSparseTree(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.leaves = [i.to_bytes(1, 'big') for i in range(10)]
        return super().setUpClass()

    def test_from_leaves(self):
        tree = sparse.SparseTree.from_leaves(self.leaves)
        assert isinstance(tree, sparse.SparseTree)
        assert hasattr(tree, 'root')
        assert type(tree.root) is bytes
        assert len(tree.root) == 32
        assert hasattr(tree, 'subtrees')
        assert type(tree.subtrees) is list
        for item in tree.subtrees:
            assert isinstance(item, sparse.SparseSubTree)

    def test_proof_e2e(self):
        tree = sparse.SparseTree.from_leaves(self.leaves)
        proof = tree.prove(self.leaves[0])
        assert type(proof) is bytes
        VM = vm.VirtualMachine(proof)
        assert VM.run(), VM.get_errors()
        assert VM.registers['return'] == tree.root, VM.registers

if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from .errors import ImplementationError, SecurityError, eruces, yert, tert
from .interfaces import VMProtocol
from enum import Enum
from hashlib import sha256
from typing import Callable


_HASH_FUNCTION = lambda input: sha256(input).digest()

def set_hash_function(hash_function: Callable[[bytes], bytes]) -> None:
    """Sets the hash function to be used by the VM. Raises
        TypeError if the hash function is not callable or
        ImplementationError if the hash function returns a non-bytes
        value when called.
    """
    tert(callable(hash_function), 'hash_function must be callable')
    try:
        output = hash_function(b'test')
        if not type(output) is bytes:
            raise ImplementationError('hash_function must return bytes when called')
        global _HASH_FUNCTION
        _HASH_FUNCTION = hash_function
    except BaseException as e:
        raise ImplementationError(f'hash_function execution failed with {e}')

def get_hash_function() -> Callable[[bytes], bytes]:
    """Returns the hash function currently being used by the VM."""
    return _HASH_FUNCTION

def hash_leaf(data: bytes) -> bytes:
    """Hashes a leaf node."""
    return get_hash_function()(b'\x00' + data)

def hash_node(left: bytes, right: bytes) -> bytes:
    """Hashes an internal node."""
    return get_hash_function()(b'\x01' + left + right)

def xor(b1: bytes, b2: bytes) -> bytes:
    """XOR two equal-length byte strings together."""
    b3 = bytearray()
    for i in range(len(b1)):
        b3.append(b1[i] ^ b2[i])

    return bytes(b3)


class OpCode(Enum):
    """The set of op codes supported by the VM."""
    load_left_hsize = 0
    load_right_hsize = 1
    hash_left = 2
    hash_right = 3
    hash_final_hsize = 4
    hash_mid = 5
    hash_leaf_left = 6
    hash_leaf_right = 7
    hash_leaf_mid = 8
    hash_leaf_bit = 9
    hash_bit = 10
    hash_final = 11
    hash_with_empty = 12
    hash_to_level = 13
    hash_to_level_hsize = 14
    hash_to_level_path = 15
    hash_left_only = 16
    hash_right_only = 17
    hash_xor_left = 18
    hash_xor_right = 19
    hash_xor_final = 20
    load_left = 40
    load_right = 41
    load_empty_left = 42
    load_empty_right = 43
    set_hsize = 50
    set_path = 51
    set_path_hsize = 52
    set_path_auto = 53
    get_path_bit = 54
    subroutine_left = 60
    subroutine_right = 61
    move_to_left = 62
    move_to_right = 63

    def __bytes__(self) -> bytes:
        return self.value.to_bytes(1, 'big')


def load_left_hsize(vm: VMProtocol):
    """Reads the size register, then reads that many bytes into the left
        register.
    """
    size = vm.get_register('size')
    left = vm.read(size)
    if vm.get_register('left'):
        eruces(vm.get_register('left') == left,
               'cannot overwrite register')
    vm.set_register('left', left)
    vm.debug('load_left_hsize', left.hex())

def load_right_hsize(vm: VMProtocol):
    """Reads the size register, then reads that many bytes into the
        right register.
    """
    size = vm.get_register('size')
    right = vm.read(size)
    if vm.get_register('right'):
        eruces(vm.get_register('right') == right,
               'cannot overwrite register')
    vm.set_register('right', right)
    vm.debug('load_right_hsize', right.hex())

def hash_left(vm: VMProtocol):
    """Puts hash(0x01 | left | right) into left register."""
    left = vm.get_register('left')
    right = vm.get_register('right')
    left = hash_node(left, right)
    vm.set_register('left', left)
    vm.set_register('right', b'')
    vm.debug('hash_left', left.hex())

def hash_right(vm: VMProtocol):
    """Puts hash(0x01 | left | right) into right register."""
    left = vm.get_register('left')
    right = vm.get_register('right')
    right = hash_node(left, right)
    vm.set_register('right', right)
    vm.set_register('left', b'')
    vm.debug('hash_right', right.hex())

def hash_final_hsize(vm: VMProtocol):
    """Reads size register, then reads that many bytes as root hash.
        Calculates hash(0x01 | left | right). Puts True in final
        register if they match and False otherwise.
    """
    size = vm.get_register('size')
    expected_root = vm.read(size)
    left = vm.get_register('left')
    right = vm.get_register('right')
    observed_root = hash_node(left, right)
    vm.set_register('final', observed_root == expected_root)
    vm.set_register('return', observed_root)
    vm.debug('hash_final_hsize', observed_root.hex())

def hash_mid(vm: VMProtocol):
    """Read left and right registers. Put hash_node(left, right) into
        return register.
    """
    left = vm.get_register('left')
    right = vm.get_register('right')
    result = hash_node(left, right)
    vm.set_register('return', result)
    vm.debug('hash_mid', result.hex())

def set_hsize(vm: VMProtocol):
    """Reads next byte, interpreting as uint8. Set the 'size' register
        to that value.
    """
    size = vm.read(1)[0]
    vm.set_register('size', size)
    vm.debug('set_hsize', size)

def load_left(vm: VMProtocol):
    """Reads next 2 bytes, interpreting as uint16. Reads that many bytes
        into the left register.
    """
    size = int.from_bytes(vm.read(2), 'big')
    left = vm.read(size)
    if vm.get_register('left'):
        eruces(vm.get_register('left') == left,
               'cannot overwrite register')
    vm.set_register('left', left)
    vm.debug('load_left', left.hex())

def load_right(vm: VMProtocol):
    """Reads next 2 bytes, interpreting as uint16. Reads that many bytes
        into the right register.
    """
    size = int.from_bytes(vm.read(2), 'big')
    right = vm.read(size)
    if vm.get_register('left'):
        eruces(vm.get_register('right') == right,
               'cannot overwrite register')
    vm.set_register('right', right)
    vm.debug('load_right', right.hex())

def hash_leaf_left(vm: VMProtocol):
    """Puts hash(0x00 | left | right) into left register."""
    left = vm.get_register('left')
    right = vm.get_register('right')
    left = hash_leaf(left + right)
    vm.set_register('left', left)
    vm.set_register('right', b'')
    vm.debug('hash_leaf_left', left.hex())

def hash_leaf_right(vm: VMProtocol):
    """Puts hash(0x00 | left | right) into right register."""
    left = vm.get_register('left')
    right = vm.get_register('right')
    right = hash_leaf(left + right)
    vm.set_register('right', right)
    vm.set_register('left', b'')
    vm.debug('hash_leaf_right', right.hex())

def hash_leaf_mid(vm: VMProtocol):
    """Read left and right registers. Put hash_leaf(left + right) into
        return register.
    """
    left = vm.get_register('left')
    right = vm.get_register('right')
    result = hash_leaf(left + right)
    vm.set_register('return', result)
    vm.debug('hash_leaf_mid', result.hex())

def hash_leaf_bit(vm: VMProtocol):
    """Read left, right, and bit registers. If bit, put hash_leaf(left +
        right) into right register; else, put hash_leaf(left + right)
        into left register.
    """
    left = vm.get_register('left')
    right = vm.get_register('right')
    bit = vm.get_register('bit')
    result = hash_leaf(left + right)
    vm.set_register('right' if bit else 'left', result)
    vm.debug('hash_leaf_bit', bit, result.hex())

def hash_bit(vm: VMProtocol):
    """Read left, right, and bit registers. If bit, put hash_node(left,
        right) into right register; else, put hash_node(left, right)
        into left register.
    """
    left = vm.get_register('left')
    right = vm.get_register('right')
    bit = vm.get_register('bit')
    result = hash_node(left, right)
    vm.set_register('right' if bit else 'left', result)
    vm.debug('hash_bit', bit, result.hex())

def hash_final(vm: VMProtocol):
    """Reads next byte, interpreting as uint8. Read that many bytes as
        root hash. Calculates hash(0x01 | left | right). Puts True in
        final register if they match and False otherwise. Puts the
        calculated final hash in the return register.
    """
    size = vm.read(1)[0]
    expected_root = vm.read(size)
    left = vm.get_register('left')
    right = vm.get_register('right')
    observed_root = hash_node(left, right)
    vm.set_register('final', observed_root == expected_root)
    vm.set_register('return', observed_root)
    vm.debug('hash_final', observed_root.hex(), expected_root.hex())

def hash_left_only(vm: VMProtocol):
    """Puts hash(left) into the left register."""
    left = vm.get_register('left')
    left = get_hash_function()(left)
    vm.set_register('left', left)
    vm.debug('hash_left_only', left.hex())

def hash_right_only(vm: VMProtocol):
    """Puts hash(right) into the right register."""
    right = vm.get_register('right')
    right = get_hash_function()(right)
    vm.set_register('right', right)
    vm.debug('hash_right_only', right.hex())

def hash_xor_left(vm: VMProtocol):
    """Puts xor(hash(left), hash(right)) into the left register. Clears
        the right register.
    """
    left = vm.get_register('left')
    right = vm.get_register('right')
    left = get_hash_function()(left)
    right = get_hash_function()(right)
    left = xor(left, right)
    vm.set_register('left', left)
    vm.set_register('right', b'')
    vm.debug('hash_xor_left', left)

def hash_xor_right(vm: VMProtocol):
    """Puts xor(hash(left), hash(right)) into the right register. Clears
        the left register.
    """
    left = vm.get_register('left')
    right = vm.get_register('right')
    left = get_hash_function()(left)
    right = get_hash_function()(right)
    right = xor(left, right)
    vm.set_register('left', b'')
    vm.set_register('right', right)
    vm.debug('hash_xor_left', right)

def hash_xor_final(vm: VMProtocol):
    """Reads next byte, interpreting as uint8. Read that many bytes as
        root hash. Calculates xor(hash(left), hash(right)). Puts True in
        final register if the root hash matches the calculated root hash
        and False otherwise. Puts the calculated final hash in the return
        register.
    """
    size = vm.read(1)[0]
    expected_root = vm.read(size)
    left = vm.get_register('left')
    right = vm.get_register('right')
    observed_root = xor(get_hash_function()(left), get_hash_function()(right))
    vm.set_register('final', observed_root == expected_root)
    vm.set_register('return', observed_root)
    vm.debug('hash_final', observed_root.hex(), expected_root.hex())

def subroutine_left(vm: VMProtocol):
    """Read 2 bytes as uint16. Read that many bytes as subroutine. Run
        the subroutine in a new VM. Subroutine must end with successful
        final_hash, then that hash will be put in the left register.
        Raises SecurityError if the subroutine did not execute
        successfully; or if the VM left register is set and the
        subroutine final register is set to something different.
    """
    size = int.from_bytes(vm.read(2), 'big')
    subroutine = vm.read(size)
    subvm = vm.__class__(subroutine)
    if not subvm.run():
        raise SecurityError("subroutine failed")
    left = vm.get_register('left')
    result = subvm.get_register('return')
    if left:
        eruces(left == result, 'cannot overwrite register')
    vm.set_register('left', result)
    vm.debug('subroutine_left', result.hex())

def subroutine_right(vm: VMProtocol):
    """Read 2 bytes as uint16. Read that many bytes as subroutine. Run
        the subroutine in a new VM. Subroutine must end with successful
        final_hash, then that hash will be put in the right register
        Raises SecurityError if the subroutine did not execute
        successfully; or if the VM right register is set and the
        subroutine final register is set to something different.
    """
    size = int.from_bytes(vm.read(2), 'big')
    subroutine = vm.read(size)
    subvm = vm.__class__(subroutine)
    if not subvm.run():
        raise SecurityError("subroutine failed")
    right = vm.get_register('right')
    result = subvm.get_register('return')
    if right:
        eruces(right == result, 'cannot overwrite register')
    vm.set_register('right', result)
    vm.debug('subroutine_right', result.hex())

def set_path(vm: VMProtocol):
    """Read next 2 bytes as uint16 size. Read size bytes into path
        register.
    """
    size = int.from_bytes(vm.read(2), 'big')
    path = vm.read(size)
    vm.debug('set_path', path.hex())
    vm.set_register('path', path)

def set_path_hsize(vm: VMProtocol):
    """Read size register. Read size bytes into path register."""
    size = vm.get_register('size')
    path = vm.read(size)
    vm.debug('set_path_hsize', path.hex())
    vm.set_register('path', path)

def set_path_auto(vm: VMProtocol):
    """If the left register is not empty, load it into the path register.
        If the right register is not empty, load it into the path
        register.
    """
    left = vm.get_register('left')
    if left:
        vm.set_register('path', left)
        vm.debug('set_path_auto', 'l', left.hex())
    else:
        right = vm.get_register('right')
        vm.set_register('path', right)
        vm.debug('set_path_auto', 'r', right.hex())

def get_path_bit(vm: VMProtocol):
    """Read next byte as uint8 index. Get the bit at the index from the
        path register and put into the bit register.
    """
    index = vm.read(1)[0]
    path = vm.get_register('path')
    byte_index = index // 8
    bit_index = index - byte_index*8
    path_byte = path[byte_index]
    path = "{0:b}".format(path_byte)
    path = "".join(["0" for _ in range(8-len(path))]) + path
    bit = path[bit_index] == "1"
    vm.set_register('bit', bit)
    vm.debug('get_path_bit', bit)

def hash_to_level(vm: VMProtocol):
    """Read next 2 bytes as uint8 from_level and to_level. Read next 2
        bytes as uint16 size. Read size bytes as bit path. Hash against
        empty node until reaching to_level, then put the result in the
        return register.
    """
    from_level = vm.read(1)[0]
    to_level = vm.read(1)[0]
    size = int.from_bytes(vm.read(2), 'big')
    path_bytes = vm.read(size)
    vm.debug('hash_to_level', 'start', path_bytes.hex())
    path_bits = []
    for i in range(size):
        path_bits.extend([
            0b10000000 & path_bytes[i] != 0,
            0b01000000 & path_bytes[i] != 0,
            0b00100000 & path_bytes[i] != 0,
            0b00010000 & path_bytes[i] != 0,
            0b00001000 & path_bytes[i] != 0,
            0b00000100 & path_bytes[i] != 0,
            0b00000010 & path_bytes[i] != 0,
            0b00000001 & path_bytes[i] != 0,
        ])

    for i in range(from_level, to_level):
        if not path_bits[i]:
            if i == 0:
                hash_leaf_left(vm)
            else:
                hash_left(vm)
            vm.insert_code(i.to_bytes(1, 'big'))
            load_empty_right(vm)
        else:
            if i == 0:
                hash_leaf_right(vm)
            else:
                hash_right(vm)
            vm.insert_code(i.to_bytes(1, 'big'))
            load_empty_left(vm)
    hash_left(vm)
    vm.set_register('return', vm.get_register('left'))
    vm.set_register('left', b'')
    vm.debug('hash_to_level', vm.get_register('return').hex())

def hash_to_level_hsize(vm: VMProtocol):
    """Read next 2 bytes as uint8 from_level and to_level. Read size
        from register. Read size bytes as bit path. Hash against empty
        node until reaching to_level, then put the result in the return
        register.
    """
    from_level = vm.read(1)[0]
    to_level = vm.read(1)[0]
    size = vm.get_register('size')
    path_bytes = vm.read(size)
    vm.debug('hash_to_level_hsize', 'start', path_bytes.hex(), increment_context=True)
    path_bits = []
    for i in range(size):
        path_bits.extend([
            0b10000000 & path_bytes[i] != 0,
            0b01000000 & path_bytes[i] != 0,
            0b00100000 & path_bytes[i] != 0,
            0b00010000 & path_bytes[i] != 0,
            0b00001000 & path_bytes[i] != 0,
            0b00000100 & path_bytes[i] != 0,
            0b00000010 & path_bytes[i] != 0,
            0b00000001 & path_bytes[i] != 0,
        ])

    for i in range(from_level, to_level):
        if not path_bits[i]:
            if i == 0:
                hash_leaf_left(vm)
            else:
                hash_left(vm)
            vm.insert_code(i.to_bytes(1, 'big'))
            load_empty_right(vm)
        else:
            if i == 0:
                hash_leaf_right(vm)
            else:
                hash_right(vm)
            vm.insert_code(i.to_bytes(1, 'big'))
            load_empty_left(vm)
    hash_left(vm)
    vm.set_register('return', vm.get_register('left'))
    vm.set_register('left', b'')
    vm.debug('hash_to_level_hsize', 'end', vm.get_register('return').hex(),
             decrement_context=True)

def hash_with_empty(vm: VMProtocol):
    """Read next byte as uint8 index. Read bit register. Load indexth
        recursively hashed empty node into the empty register (left or
        right), then hash_node and put the result in the register
        indicated by bit.
    """
    index = vm.read(1)[0]
    bit = vm.get_register('bit')
    left = vm.get_register('left')
    right = vm.get_register('right')
    empty = get_empty_hash(index)
    if left:
        right = empty
    else:
        left = empty
    result = hash_node(left, right)

    if not bit:
        vm.set_register('left', result)
        vm.set_register('right', b'')
    else:
        vm.set_register('right', result)
        vm.set_register('left', b'')
    vm.debug('hash_with_empty', index, bit, result.hex())

def hash_to_level_path(vm: VMProtocol):
    """Read the next 2 bytes as uint8 from_level and to_level. Read the
        path register. Hash against recursively hashed empty node until
        reaching to_level starting at from_level, then put the result in
        the return register.
    """
    from_level = vm.read(1)[0]
    to_level = vm.read(1)[0]
    path = vm.get_register('path')
    vm.debug('hash_to_level_path', 'start', path.hex(), increment_context=True)
    for i in range(from_level, to_level):
        vm.insert_code((i+1).to_bytes(1, 'big'))
        get_path_bit(vm)
        vm.insert_code(i.to_bytes(1, 'big'))
        hash_with_empty(vm)
    left = vm.get_register('left')
    right = vm.get_register('right')
    result = left or right
    vm.set_register('return', result)
    vm.set_register('left', b'')
    vm.set_register('right', b'')
    vm.debug('hash_to_level_path', 'end', result.hex(), decrement_context=True)

_EMPTY_HASHES = []

def compute_empty_hashes(empty_leaf: bytes = b''):
    while len(_EMPTY_HASHES):
        _EMPTY_HASHES.pop()
    _EMPTY_HASHES.append(get_hash_function()(b'\x00' + empty_leaf))
    for i in range(0, 255):
        preimage = _EMPTY_HASHES[i]
        _EMPTY_HASHES.append(get_hash_function()(b'\x01' + preimage + preimage))

def get_empty_hash(level: int) -> bytes:
    if not _EMPTY_HASHES:
        compute_empty_hashes()
    return _EMPTY_HASHES[level]

def load_empty_left(vm: VMProtocol):
    """Reads the next byte, interpreting as uint8 level. Loads the
        recursively hashed empty leaf for that level of the Sparse
        Merkle Tree into the left register.
    """
    level = vm.read(1)[0]
    hash = get_empty_hash(level)
    left = vm.get_register('left')
    if left:
        eruces(left == hash, 'cannot overwrite register')
    vm.set_register('left', hash)
    vm.debug('load_empty_left', hash.hex())

def load_empty_right(vm: VMProtocol):
    """Reads the next byte, interpreting as uint8 level. Loads the
        recursively hashed empty leaf for that level of the Sparse
        Merkle Tree into the right register.
    """
    level = vm.read(1)[0]
    hash = get_empty_hash(level)
    right = vm.get_register('right')
    if right:
        eruces(right == hash, 'cannot overwrite register')
    vm.set_register('right', hash)
    vm.debug('load_empty_right', hash.hex())

def move_to_left(vm: VMProtocol):
    """Moves a value from the return register to the left register."""
    result = vm.get_register('return')
    left = vm.get_register('left')
    eruces(left == b'', 'cannot overwrite register')
    vm.set_register('left', result)
    vm.set_register('return', b'')
    vm.debug('move_to_left', result.hex())

def move_to_right(vm: VMProtocol):
    """Moves a value from the return register to the right register."""
    result = vm.get_register('return')
    right = vm.get_register('right')
    eruces(right == b'', 'cannot overwrite register')
    vm.set_register('right', result)
    vm.set_register('return', b'')
    vm.debug('move_to_right', result.hex())


instruction_set = {
    OpCode.load_left_hsize: load_left_hsize,
    OpCode.load_right_hsize: load_right_hsize,
    OpCode.hash_left: hash_left,
    OpCode.hash_right: hash_right,
    OpCode.hash_final_hsize: hash_final_hsize,
    OpCode.hash_mid: hash_mid,
    OpCode.hash_leaf_left: hash_leaf_left,
    OpCode.hash_leaf_right: hash_leaf_right,
    OpCode.hash_leaf_mid: hash_leaf_mid,
    OpCode.hash_leaf_bit: hash_leaf_bit,
    OpCode.hash_bit: hash_bit,
    OpCode.hash_final: hash_final,
    OpCode.hash_with_empty: hash_with_empty,
    OpCode.hash_to_level: hash_to_level,
    OpCode.hash_to_level_hsize: hash_to_level_hsize,
    OpCode.hash_to_level_path: hash_to_level_path,
    OpCode.hash_left_only: hash_left_only,
    OpCode.hash_right_only: hash_right_only,
    OpCode.hash_xor_left: hash_xor_left,
    OpCode.hash_xor_left: hash_xor_left,
    OpCode.hash_xor_right: hash_xor_right,
    OpCode.hash_xor_final: hash_xor_final,
    OpCode.load_left: load_left,
    OpCode.load_right: load_right,
    OpCode.load_empty_left: load_empty_left,
    OpCode.load_empty_right: load_empty_right,
    OpCode.set_hsize: set_hsize,
    OpCode.set_path: set_path,
    OpCode.set_path_hsize: set_path_hsize,
    OpCode.set_path_auto: set_path_auto,
    OpCode.get_path_bit: get_path_bit,
    OpCode.subroutine_left: subroutine_left,
    OpCode.subroutine_right: subroutine_right,
    OpCode.move_to_left: move_to_left,
    OpCode.move_to_right: move_to_right,
}


class VirtualMachine:
    program: bytes
    pointer: int
    instruction_set: dict
    registers: dict[str, bytes|bool|int|None|list]

    def __init__(self, program: bytes = b'', pointer: int = 0,
                 instruction_set: dict[OpCode, Callable] = instruction_set,
                 debug: bool = False) -> None:
        """Initialize the VM with the given program, pointer, instruction
            set, and debug flag.
        """
        self.program = program
        self.pointer = pointer
        self.instruction_set = instruction_set
        self.registers = {
            'left': b'',
            'right': b'',
            'path': b'',
            'bit': False,
            'final': False,
            'size': 32,
            'return': None,
            'errors': [],
        }
        self._debug_enabled = debug
        self._debug_context = 0

    def run(self) -> bool:
        """Runs the program. Returns True if the proof was verified
            successfully and False otherwise.
        """
        while self.pointer < len(self.program):
            if not self.execute_next():
                return False

        return self.registers['final']

    def read(self, count: int = 1) -> bytes:
        """Reads count bytes from the program, advances the pointer, and
            returns the read bytes.
        """
        result = self.program[self.pointer:self.pointer+count]
        self.pointer += count
        return result

    def execute_next(self) -> bool:
        """Runs the next instruction. Returns True if it ran
            successfully and False otherwise.
        """
        try:
            op = self.read(1)[0]
            code = OpCode(op)
            self.instruction_set[code](self)
            return True
        except BaseException as e:
            self.registers['errors'].append(e)
            return False

    def reset(self) -> VirtualMachine:
        """Resets the instance and returns self."""
        self.pointer = 0
        self.registers = {
            'left': b'',
            'right': b'',
            'path': b'',
            'bit': False,
            'final': False,
            'size': 32,
            'return': None,
            'errors': [],
        }
        return self

    def load_program(self, program: bytes = b'', pointer: int = 0) -> None:
        """Loads the supplied program and resets the instruction pointer."""
        self.program = program
        self.pointer = pointer

    def insert_code(self, code: bytes) -> None:
        """Inserts code at the current pointer."""
        self.program = self.program[:self.pointer] + code + self.program[self.pointer:]

    def set_register(self, name: str, value: bytes|int|bool|None) -> None:
        """Sets the specified register to the given value."""
        self.registers[name] = value

    def get_register(self, name: str) -> bytes|int|bool|None:
        """Returns the value of the specified register."""
        return self.registers[name]

    def has_completed(self) -> bool:
        """Returns True if the instruction pointer is >= the length of
            the loaded program.
        """
        return self.pointer >= len(self.program)

    def get_errors(self) -> list[BaseException]:
        """Returns any errors that occurred during execution."""
        return self.get_register('errors') or []

    def debug(self, *parts, increment_context: bool = False,
              decrement_context: bool = False) -> None:
        """If debug is enabled, add a debug trace."""
        if decrement_context:
            self._debug_context -= 1
            if self._debug_context < 0:
                self._debug_context = 0

        if self._debug_enabled:
            if self._debug_context:
                print(
                    ''.join(['\t' for _ in range (self._debug_context)]),
                    *parts
                )
            else:
                print(*parts)

        if increment_context:
            self._debug_context += 1


def compile(*symbols: OpCode|bytes|int|tuple[OpCode|bytes|int,]) -> bytes:
    """Compiles a list of OpCode, bytes, and ints into byte code.
        Raises SyntaxError for invalid VM code syntax. Raises TypeError
        for invalid symbols.
    """
    flattened = []
    for symbol, index in zip(symbols, range(len(symbols))):
        tert(type(symbol) in (OpCode, bytes, int, tuple),
            f"Symbol at {index}: type {type(symbol)} not supported")
        if type(symbol) is tuple:
            tert(all([type(item) in (OpCode, bytes, int) for item in symbol]),
                 f"Symbol at {index}: tuple must contain OpCode, bytes, and/or int")
            flattened.extend(symbol)
        else:
            flattened.append(symbol)

    index = 0
    code = b''
    while index < len(flattened):
        to_add, advance = _compile_next(index, flattened)
        code += to_add
        index += advance

    return code

_advance_ = {
    'op': (
        OpCode.hash_mid,
        OpCode.hash_left,
        OpCode.hash_right,
        OpCode.hash_leaf_left,
        OpCode.hash_leaf_right,
        OpCode.hash_leaf_mid,
        OpCode.hash_leaf_bit,
        OpCode.hash_bit,
        OpCode.hash_left_only,
        OpCode.hash_right_only,
        OpCode.hash_xor_left,
        OpCode.hash_xor_right,
        OpCode.set_path_auto,
    ),
    'op bytes': (
        OpCode.load_left_hsize,
        OpCode.load_right_hsize,
        OpCode.hash_final_hsize,
        OpCode.set_path_hsize,
    ),
    'op u8 bytes': (
        OpCode.hash_final,
        OpCode.hash_xor_final,
    ),
    'op u16 bytes': (
        OpCode.load_left,
        OpCode.load_right,
        OpCode.subroutine_left,
        OpCode.subroutine_right,
        OpCode.set_path,
    ),
    'op u8 u8 u16 bytes': (
        OpCode.hash_to_level,
    ),
    'op u8 u8 bytes': (
        OpCode.hash_to_level_hsize,
    ),
    'op u8': (
        OpCode.set_hsize,
        OpCode.get_path_bit,
        OpCode.hash_with_empty,
        OpCode.load_empty_left,
        OpCode.load_empty_right,
    ),
    'op u8 u8': (
        OpCode.hash_to_level_path,
    )
}

def _compile_next(index: int, symbols: list[OpCode|bytes|int,]) -> tuple[bytes, int]:
    """Compiles the next op into byte code using remaining symbols.
        Returns the byte code and the number of symbols to advance.
        Raises SyntaxError if the symbol at the index is not an OpCode
        or if the symbols did not contain valid parameters for the
        OpCode.
    """
    yert(type(symbols[index]) is OpCode,
        f"Symbol at {index}: expected OpCode element.")

    def check_param_count(symbols: list[OpCode|bytes|int], index: int, count: int, params: str):
        yert(len(symbols[index:]) >= count+1,
             f"Symbol at {index}: expected {'params' if count>1 else 'param'} {params}")

    def check_u8(symbols: list[OpCode|bytes|int], index: int, offset: int):
        yert(type(symbols[index+offset]) is int,
             f"Symbol at {index+offset} (after {op.name}): expected int; " +
             f"{type(symbols[index+offset])} not supported")
        yert(0 <= symbols[index+offset] < 256,
             f"Symbol at {index+offset} (after {op.name}): " +
             f"expected 0<={symbols[index+offset]}<=255")

    def check_bytes_255(symbols: list[OpCode|bytes|int], index: int, offset: int):
        yert(type(symbols[index+offset]) is bytes,
             f"Symbol at {index+offset} (after {op.name} {symbols[index+1]} " +
             f"{symbols[index+offset]}): expected bytes up to 255 length; " +
             f"{type(symbols[index+offset])} is not supported")
        yert(len(symbols[index+offset]) < 256,
             f"Symbol at {index+offset} (after {op.name} {symbols[index+1]} " +
             f"{symbols[index+offset]}): expected bytes up to 255 length; " +
             f"{len(symbols[index+offset])} is too large")

    def check_bytes_65535(symbols: list[OpCode|bytes|int], index: int, offset: int):
        yert(type(symbols[index+offset]) is bytes,
             f"Symbol at {index+offset} (after {op.name} {symbols[index+1]}): " +
             f"expected bytes up to 65535 length; " +
             f"{type(symbols[index+offset])} is not supported")
        yert(len(symbols[index+offset]) < 2**16,
             f"Symbol at {index+offset} (after {op.name} {symbols[index+1]} " +
             f"{symbols[index+offset]}): expected bytes up to 65535 length; " +
             f"{len(symbols[index+offset])} is too large")

    op = symbols[index]
    code = bytes(op)

    if op in _advance_['op bytes']:
        check_param_count(symbols, index, 1, "bytes")
        check_bytes_65535(symbols, index, 1)
        code += symbols[index+1]
        return (code, 2)

    if op in _advance_['op u8 bytes']:
        check_param_count(symbols, index, 1, "bytes(len<=255)")
        check_bytes_255(symbols, index, 1)
        code += len(symbols[index+1]).to_bytes(1, 'big')
        code += symbols[index+1]
        return (code, 2)

    if op in _advance_['op u16 bytes']:
        check_param_count(symbols, index, 1, "bytes(len<=65535)")
        check_bytes_65535(symbols, index, 1)
        code += len(symbols[index+1]).to_bytes(2, 'big')
        code += symbols[index+1]
        return (code, 2)

    if op in _advance_['op u8 u8 u16 bytes']:
        check_param_count(symbols, index, 3, "uint8 uint8 bytes(len<=65535)")
        yert(len(symbols[index:]) >= 3,
             f"Symbol at {index}: expected params uint8 uint8 bytes(len<=65535)")
        check_u8(symbols, index, 1)
        check_u8(symbols, index, 2)
        check_bytes_65535(symbols, index, 3)
        code += symbols[index+1].to_bytes(1, 'big')
        code += symbols[index+2].to_bytes(1, 'big')
        code += len(symbols[index+3]).to_bytes(2, 'big')
        code += symbols[index+3]
        return (code, 4)

    if op in _advance_['op u8 u8 bytes']:
        check_param_count(symbols, index, 3, "uint8 uint8 bytes(len<=65535)")
        check_u8(symbols, index, 1)
        check_u8(symbols, index, 2)
        check_bytes_65535(symbols, index, 3)
        code += symbols[index+1].to_bytes(1, 'big')
        code += symbols[index+2].to_bytes(1, 'big')
        code += symbols[index+3]
        return (code, 4)

    if op in _advance_['op u8 u8']:
        check_param_count(symbols, index, 2, "uint8 uint8")
        check_u8(symbols, index, 1)
        check_u8(symbols, index, 2)
        code += symbols[index+1].to_bytes(1, 'big')
        code += symbols[index+2].to_bytes(1, 'big')
        return (code, 3)

    if op in _advance_['op u8']:
        check_param_count(symbols, index, 1, "uint8")
        check_u8(symbols, index, 1)
        code += symbols[index+1].to_bytes(1, 'big')
        return (code, 2)

    return (code, 1)


def decompile(code: bytes) -> list[OpCode|bytes|int,]:
    """Decompiles bytecode into a list of OpCode, bytes, and ints."""
    proof = []
    index = 0
    hsize = 32
    while index < len(code):
        op = OpCode(code[index])
        proof.append(op)
        index += 1

        if op is OpCode.set_hsize:
            hsize = code[index]
            proof.append(hsize)
            index += 1
        elif op in _advance_['op bytes']:
            proof.append(code[index:index+hsize])
            index += hsize
        elif op in _advance_['op u8 bytes']:
            size = code[index]
            proof.append(code[index+1:index+1+size])
            index += 1 + size
        elif op in _advance_['op u16 bytes']:
            size = int.from_bytes(code[index:index+2], 'big')
            proof.append(code[index+2:index+2+size])
            index += 2 + size
        elif op in _advance_['op u8 u8 bytes']:
            first = code[index]
            second = code[index+1]
            proof.extend([
                first,
                second,
                code[index+2:index+2+hsize]
            ])
            index += 2 + hsize
        elif op in _advance_['op u8 u8']:
            first = code[index]
            second = code[index+1]
            proof.extend([
                first,
                second,
            ])
            index += 2
        elif op in _advance_['op u8']:
            proof.append(code[index])
            index += 1
    return proof

def adapt_legacy_proof(proof: list[bytes], hash_size: int = 32) -> bytes:
    """Adapts a proof from the legacy code into bytecode. Raises
        `TypeError` or `ValueError` upon invalid input.
    """
    tert(all([type(step) is bytes for step in proof]),
           'proof must be bytes or list of bytes')
    return OpCode.set_hsize.value.to_bytes(1, 'big') + \
        (hash_size).to_bytes(1, 'big') + b''.join(proof)



from __future__ import annotations
from enum import Enum
from typing import Any, Callable, Hashable, Protocol, runtime_checkable


@runtime_checkable
class VMProtocol(Protocol):
    def __init__(self, program: bytes = b'', pointer: int = 0,
                 instruction_set: dict[type[Enum], Callable] = {},
                 debug: bool = False) -> None:
        """Initialize a VM with the given program and instruction_set."""
        ...

    def run(self) -> Any:
        """Run the loaded program. Return the accumulated value."""
        ...

    def read(self, count: int = 1) -> bytes:
        """Read the given number of bytes."""
        ...

    def execute_next(self) -> bool:
        """Execute the next operation. Return True if successful and
            False otherwise.
        """
        ...

    def reset(self) -> VMProtocol:
        """Resets the virtual machine and returns self."""
        ...

    def load_program(self, program: bytes = b'', pointer: int = 0) -> None:
        """Loads the supplied program and resets the instruction pointer."""
        ...

    def insert_code(self, code: bytes) -> None:
        """Inserts code at the current pointer."""
        ...

    def set_register(self, name: Hashable, value: Any) -> None:
        """Sets the specified register to the given value."""
        ...

    def get_register(self, name: Hashable) -> Any:
        """Returns the value of the specified register."""
        ...

    def has_completed(self) -> bool:
        """Returns True if the VM has completed running the program."""
        ...

    def get_errors(self) -> list[BaseException]:
        """Returns any errors that occurred during execution."""
        ...

    def debug(self, *parts, increment_context: bool = False,
              decrement_context: bool = False) -> None:
        """If debug is enabled, add a debug trace."""
        ...


@runtime_checkable
class PackableProtocol(Protocol):
    def pack(self) -> bytes:
        """Packs the instance into bytes."""
        ...

    @classmethod
    def unpack(cls, data: bytes, /, *, inject: dict = {}) -> PackableProtocol:
        """Unpacks an instance from bytes. Must accept dependency
            injection to unpack other Packable types.
        """
        ...
